from bespin.errors import BadS3Bucket, BespinError
from bespin import helpers as hp

from six.moves.urllib.parse import urlparse
from contextlib import contextmanager
from collections import namedtuple
from filechunkio import FileChunkIO

from datetime import datetime
import humanize
import logging
import boto.s3
import boto
import sys
import os

from boto.s3.key import Key

log = logging.getLogger("bespin.amazon.s3")

S3Location = namedtuple("S3Location", ["bucket", "key", "full"])

class S3(object):

    def __init__(self, region="ap-southeast-2"):
        self.region = region

    @hp.memoized_property
    def conn(self):
        log.info("Using region [%s] for s3", self.region)
        return boto.s3.connect_to_region(self.region)

    def s3_location(self, value):
        """Return us an s3 location record type"""
        info = urlparse(value)
        if info.scheme != "s3":
            raise ValueError("S3 location must be a valid s3 url\tgot={0}".format(value))

        bucket = info.netloc
        if not bucket:
            raise ValueError("S3 location must be a valid s3 url\tgot={0}".format(value))

        key = info.path
        return S3Location(bucket, key, value)

    def copy_key(self, frm, to):
        """Copy a key from <frm> to <to>"""
        frm_location = self.s3_location(frm)
        to_location = self.s3_location(to)
        bucket = self.get_bucket(to_location.bucket)
        log.info("Copying %s to %s", frm, to)
        bucket.copy_key(to.key, frm_location.bucket, frm_location.key)

    def wait_for(self, bucket, key, timeout, start=None):
        if start is None:
            start = datetime.utcnow()

        log.info("Looking for key with last_modified greater than %s", start)
        for _ in hp.until(timeout=timeout, step=5):
            try:
                bucket_obj = self.get_bucket(bucket)
            except BadS3Bucket as error:
                log.error(error)
                continue

            if key == '/':
                log.info("The bucket exists! and that is all we are looking for")
                return

            k = Key(bucket_obj)
            k.key = key

            try:
                k.read()
            except boto.exception.S3ResponseError as error:
                if error.status == 404:
                    log.info("Key doesn't exist yet\tbucket=%s\tkey=%s", bucket_obj.name, key)
                    continue
                else:
                    log.error(error)
                    continue

            last_modified = k.last_modified
            log.info("Found key in the bucket\tbucket=%s\tkey=%s\tlast_modified=%s", bucket_obj.name, key, last_modified)

            date = datetime.strptime(last_modified, "%a, %d %b %Y %H:%M:%S GMT")
            if date > start:
                log.info("Found key and it's newer than our start time!")
                return
            else:
                log.info("Found key but it's older than our start time, hasn't been updated yet")

        raise BespinError("Couldn't find the s3 key with a newer last modified")

    @contextmanager
    def a_multipart_upload(self, bucket, key):
        mp = None
        try:
            mp = bucket.initiate_multipart_upload(key)
            yield mp
        except:
            if mp:
                mp.cancel_upload()
            raise
        else:
            if mp:
                mp.complete_upload()

    def get_bucket(self, bucket_name):
        try:
            return self.conn.get_bucket(bucket_name)
        except boto.exception.S3ResponseError as error:
            if error.status in (404, 403):
                raise BadS3Bucket("Bucket either doesn't exist or isn't available to you", name=bucket_name)
            else:
                raise

    def list_keys_from_s3_path(self, query_path):
        query = self.s3_location(query_path)
        bucket = self.get_bucket(query.bucket)
        return bucket.list(prefix=query.key[1:])

    def delete_key_from_s3(self, key, dry_run):
        if dry_run:
            print("Would delete {0}".format(key.name))
        else:
            print("Deleted {0}".format(key.name))
            key.delete()

    def upload_file_to_s3(self, source_filename, destination_path):
        source_size = os.stat(source_filename).st_size
        if source_size > 30000000:
            self.upload_file_to_s3_in_parts(source_filename, destination_path)
        else:
            self.upload_file_to_s3_as_single(source_filename, destination_path)

    def upload_file_to_s3_in_parts(self, source_filename, destination_path):
        source_file = open(source_filename, 'rb')
        destination_file = self.s3_location(destination_path)

        source = os.path.abspath(source_file.name)
        source_size = os.stat(source).st_size
        log.info("Uploading from %s (%s) to %s in parts", source, humanize.naturalsize(source_size), destination_file.full)

        bucket = self.get_bucket(destination_file.bucket)

        try:
            with self.a_multipart_upload(bucket, destination_file.key) as mp:
                for chunk, offset, length in self.determine_chunks(source_size, min_chunk=5242881):
                    with FileChunkIO(source, 'r', offset=offset, bytes=length) as fp:
                        log.info("Uploading chunk %s (%s)", chunk+1, humanize.naturalsize(length))
                        mp.upload_part_from_file(fp, part_num=chunk+1)

        except boto.exception.S3ResponseError as error:
            if error.status is 403:
                log.error("Seems you are unable to edit this location :(")
                sys.exit(1)
            else:
                raise

        log.info("Finished uploading")

    def determine_chunks(self, total_size, min_chunk=5242881):
        offset = 0
        offsets = []
        while offset < total_size:
            if total_size - offset < min_chunk:
                break
            offsets.append(offset)
            offset += min_chunk
        offsets.append(total_size)
        log.info("Broken up into %s chunks", len(offsets) - 1)

        for index, offset in enumerate(offsets[:-1]):
            nxt = offsets[index+1]
            yield index, offset, nxt - offset

    def upload_file_to_s3_as_single(self, source_filename, destination_path):
        destination_file = self.s3_location(destination_path)

        source = os.path.abspath(source_filename)
        source_size = os.stat(source).st_size
        log.info("Uploading from %s (%s) to %s in one go", source, humanize.naturalsize(source_size), destination_file.full)

        bucket = self.get_bucket(destination_file.bucket)
        key = Key(bucket)
        key.name = destination_file.key
        key.set_contents_from_filename(source_filename, policy='authenticated-read')
        key.close()



from bespin.amazon.cloudformation import Cloudformation
from bespin.helpers import memoized_property
from bespin.errors import BespinError, ProgrammerError
from bespin.amazon.ec2 import EC2
from bespin.amazon.sqs import SQS
from bespin.amazon.kms import KMS
from bespin.amazon.s3 import S3
from bespin import VERSION

from input_algorithms.spec_base import NotSpecified
import logging
import botocore
import boto3
import os

log = logging.getLogger("bespin.amazon.credentials")

class Credentials(object):
    def __init__(self, region, account_id, assume_role):
        self.region = region
        self.account_id = account_id
        self.assume_role = assume_role
        self.session = None
        self.clouds = {}

    def verify_creds(self):
        """Make sure our current credentials are for this account and set self.connection"""
        if getattr(self, "_verified", None):
            return

        if self.assume_role is not NotSpecified:
            self.assume()
            self._verified = True
            return

        log.info("Verifying amazon credentials")
        try:
            self.session = boto3.session.Session(region_name=self.region)
            amazon_account_id = self.session.client('sts').get_caller_identity().get('Account')
            if int(self.account_id) != int(amazon_account_id):
                raise BespinError("Please use credentials for the right account", expect=self.account_id, got=amazon_account_id)
            self._verified = True
        except botocore.exceptions.NoCredentialsError:
            raise BespinError("Export AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY before running this script (your aws credentials)")
        except botocore.exceptions.ClientError as error:
            raise BespinError("Couldn't determine what account your credentials are from", error=error.message)

        if self.session is None or self.session.region_name != self.region:
            raise ProgrammerError("botocore.session created in incorrect region")

    def account_role_arn(self, role, partition='aws'):
        """Return full ARN for a role within the current account"""
        if not role or role.startswith("arn:aws"):
            return role
        if not role.startswith("role/"):
            role = "role/" + role
        return "arn:{0}:iam::{1}:{2}".format(partition, self.account_id, role)

    def assume(self):
        assumed_role = self.account_role_arn(self.assume_role)
        log.info("Assuming role as %s", assumed_role)

        for name in ['AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY', 'AWS_SECURITY_TOKEN', 'AWS_SESSION_TOKEN']:
            if name in os.environ and not os.environ[name]:
                del os.environ[name]

        try:
            conn = boto3.client('sts', region_name=self.region)
            session_name = "{1}@bespin{0}".format(VERSION, os.environ.get("USER", "<unknown_user>"))
            response = conn.assume_role(RoleArn=assumed_role, RoleSessionName=session_name)

            role = response['AssumedRoleUser']
            creds = response['Credentials']
            log.info("Assumed role (%s)", role['Arn'])
            self.session = boto3.session.Session(
                    aws_access_key_id=creds['AccessKeyId'],
                    aws_secret_access_key=creds['SecretAccessKey'],
                    aws_session_token=creds['SessionToken'],
                    region_name=self.region
            )

            os.environ['AWS_ACCESS_KEY_ID'] = creds["AccessKeyId"]
            os.environ['AWS_SECRET_ACCESS_KEY'] = creds["SecretAccessKey"]
            os.environ['AWS_SECURITY_TOKEN'] = creds["SessionToken"]
            os.environ['AWS_SESSION_TOKEN'] = creds["SessionToken"]
        except botocore.exceptions.NoCredentialsError:
            raise BespinError("Export AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY before running this script (your aws credentials)")
        except botocore.exceptions.ClientError as error:
            raise BespinError("Unable to assume role", error=error.message)

    @memoized_property
    def s3(self):
        self.verify_creds()
        return S3(self.region)

    @memoized_property
    def ec2(self):
        self.verify_creds()
        return EC2(self.region)

    @memoized_property
    def sqs(self):
        self.verify_creds()
        return SQS(self.region)

    @memoized_property
    def kms(self):
        self.verify_creds()
        return KMS(self.region)

    @memoized_property
    def iam(self):
        self.verify_creds()
        log.info("Using region [%s] for iam", self.region)
        return self.session.client('iam', region_name=self.region)

    def cloudformation(self, stack_name):
        self.verify_creds()
        if stack_name not in self.clouds:
            self.clouds[stack_name] = Cloudformation(stack_name, self.region)
        return self.clouds[stack_name]




"""
Custom specifications for the different types of stack options.

The idea is that these understand the conditions around representation of the
options.
"""

from bespin.option_spec.stack_objs import (
      StaticVariable, DynamicVariable, Environment, Skipper, S3Address
    , UltraDNSSite, UltraDNSProvider
    )
from bespin.option_spec.artifact_objs import ArtifactCommand
from bespin.option_spec.artifact_objs import ArtifactPath
from bespin.formatter import MergedOptionStringFormatter
from bespin.errors import BadSpecValue, BadConfiguration
from bespin.helpers import memoized_property

from input_algorithms.many_item_spec import many_item_formatted_spec
from input_algorithms.spec_base import NotSpecified, Spec
from input_algorithms import spec_base as sb
from input_algorithms import validators
from six.moves.urllib.parse import urlparse
import logging

log = logging.getLogger("bespin.option_spec.stack_specs")

class var_spec(many_item_formatted_spec):
    value_name = "Variable"
    specs = [sb.or_spec(sb.string_or_int_as_string_spec(), sb.listof(sb.string_or_int_as_string_spec()))]
    optional_specs = [sb.string_or_int_as_string_spec()]
    formatter = MergedOptionStringFormatter
    seperators = "|"

    def create_result(self, variable, variable_value, meta, val, dividers):
        if variable_value is NotSpecified:
            return StaticVariable(variable)
        else:
            stack = variable
            return DynamicVariable(stack, variable_value, meta.everything["bespin"])

class artifact_path_spec(many_item_formatted_spec):
    value_name = "Artifact Path"
    specs = [sb.string_spec(), sb.string_spec()]
    creates = ArtifactPath
    formatter = MergedOptionStringFormatter

    def create_result(self, host_path, artifact_path, meta, val, dividers):
        return ArtifactPath(host_path, artifact_path)

class env_spec(many_item_formatted_spec):
    value_name = "Environment Variable"
    seperators = [':', '=']

    specs = [sb.string_spec()]
    creates = Environment
    optional_specs = [sb.string_or_int_as_string_spec()]
    formatter = MergedOptionStringFormatter

    def create_result(self, env_name, other_val, meta, val, dividers):
        """Set default_val and set_val depending on the seperator"""
        args = [env_name]
        if other_val is NotSpecified:
            other_val = None
        if not dividers:
            args.extend([None, None])
        elif dividers[0] == ':':
            args.extend([other_val, None])
        elif dividers[0] == '=':
            args.extend([None, other_val])
        return Environment(*args)

class skipper_spec(many_item_formatted_spec):
    value_name = "Skip specification"
    spec = lambda: sb.delayed(var_spec())
    creates = Skipper
    specs = [spec(), spec()]

    def create_result(self, var1, var2, meta, val, dividers):
        return Skipper(var1, var2)

class s3_address(many_item_formatted_spec):
    value_name = "s3 address"
    specs = [sb.string_spec()]
    optional_specs = [sb.integer_spec()]
    creates = S3Address
    seperators = None
    formatter = MergedOptionStringFormatter

    def create_result(self, address, timeout, meta, val, dividers):
        if timeout is NotSpecified:
            timeout = 600

        options = urlparse(address)
        if options.scheme != "s3":
            raise BadSpecValue("Not a valid s3 address", meta=meta, got=val)
        if not options.netloc:
            path = ''
            domain = options.path
        else:
            path = options.path
            domain = options.netloc

        if not path.startswith('/'):
            path  = '/'
        return S3Address(domain, path, timeout)

class dns_spec(Spec):
    def setup(self, spec):
        self.spec = spec

    def normalise_filled(self, meta, val):
        meta.everything = meta.everything.wrapped()
        meta.everything["__dns_vars__"] = val["vars"].as_dict()
        return self.spec.normalise(meta, val)

class dns_site_spec(Spec):
    def normalise_filled(self, meta, val):
        log.info("Normalising dns site %s", meta.path)
        val = sb.dictionary_spec().normalise(meta, val)
        provider = val["provider"]
        available = meta.everything["stacks"][meta.everything["__stack_name__"]]["dns"]["providers"]
        if provider not in available.keys():
            raise BadConfiguration("Specified provider isn't defined in {dns.providers}", available=list(available.keys()), wanted=provider, meta=meta)

        val["provider"] = lambda: meta.everything["stacks"][meta.everything["__stack_name__"]]["dns"]["providers"][provider]
        if available[provider]["provider_type"] == "ultradns":
            return self.ultradns_site_spec(val).normalise(meta, val)
        else:
            raise BadConfiguration("Unknown dns provider type", available=["ultradns"], wanted=val["provider"].provider_type, meta=meta)

    def ultradns_site_spec(self, this):
        formatted_string = sb.formatted(sb.string_spec(), formatter=MergedOptionStringFormatter)
        return sb.create_spec(UltraDNSSite
            , name = sb.formatted(sb.overridden("{_key_name_1}"), formatter=MergedOptionStringFormatter)
            , ttl = sb.optional_spec(sb.integer_spec())
            , provider = sb.any_spec()
            , record_type = sb.required(formatted_string)
            , zone = sb.required(formatted_string)
            , domain = sb.required(formatted_string)
            , environments = sb.required(self.dns_environment_spec(this))
            )

    def dns_environment_spec(self, this):
        formatted_string = sb.formatted(sb.string_spec(), formatter=MergedOptionStringFormatter)
        class spec(Spec):
            def normalise_filled(s, meta, val):
                meta.everything = meta.everything.wrapped()
                meta.everything["__site_environments__"] = this["environments"].as_dict()
                spec = sb.dictof(sb.string_spec(), sb.listof(formatted_string))
                return spec.normalise(meta, val.as_dict())
        return spec()

class dns_provider_spec(Spec):
    def normalise_filled(self, meta, val):
        val = sb.dictionary_spec().normalise(meta, val)
        provider_type = val["provider_type"]
        available = ["ultradns"]
        if provider_type not in available:
            raise BadConfiguration("Specified provider type isn't supported", supported=available, wanted=provider_type, meta=meta)

        if provider_type == "ultradns":
            return self.ultradns_provider_spec.normalise(meta, val)

    @memoized_property
    def ultradns_provider_spec(self):
        return sb.create_spec(UltraDNSProvider
            , name = sb.formatted(sb.overridden("{_key_name_1}"), formatter=MergedOptionStringFormatter)
            , provider_type = sb.required(sb.string_spec())
            , username = sb.required(formatted_string)
            , password = sb.required(formatted_string)
            )

formatted_string = sb.formatted(sb.string_spec(), formatter=MergedOptionStringFormatter)

artifact_command_spec = lambda : sb.create_spec(ArtifactCommand
    , copy = sb.listof(artifact_path_spec())
    , modify = sb.dictof(sb.string_spec(), sb.set_options(append=sb.listof(formatted_string)))
    , command = sb.listof(formatted_string)
    , timeout = sb.defaulted(sb.integer_spec(), 600)
    , temp_dir = sb.defaulted(formatted_string, None)
    , add_into_tar = sb.listof(artifact_path_spec())
    )

params_json_spec = lambda: sb.listof(sb.set_options(
      ParameterKey = sb.required(sb.any_spec())
    , ParameterValue = sb.required(sb.any_spec())
    ))

params_yaml_spec = lambda: sb.dictof(sb.string_spec(), sb.formatted(sb.string_spec(), formatter=MergedOptionStringFormatter))

stack_json_spec = lambda: sb.set_options(
      Resources = sb.required(sb.dictof(sb.string_spec(), sb.set_options(Type=sb.required(sb.string_spec()), Properties=sb.optional_spec(sb.dictionary_spec()))))
    , Parameters = sb.optional_spec(sb.dictof(sb.string_spec(), sb.dictionary_spec()))
    , Outputs = sb.optional_spec(sb.dictof(sb.string_spec(), sb.dictionary_spec()))
    )

class policy_set_options(sb.set_options):
    """
    Strip ``NotSpecified`` values from normalised dictionary
    and ensure only one of the 'Not' options are specified
    """
    def normalise_filled(self, meta, val):
        result = super(policy_set_options, self).normalise_filled(meta, val)
        val = dict((k, v) for k, v in result.iteritems() if v is not NotSpecified)
        return sb.apply_validators(meta, val, [
                validators.has_only_one_of(["Action", "NotAction"]),
                validators.has_only_one_of(["Resource", "NotResource"])
            ])

# http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/protect-stack-resources.html#stack-policy-reference
policy_json_spec = lambda: sb.set_options(
      Statement = sb.listof(policy_set_options(
          Effect = sb.string_choice_spec(choices=["Deny", "Allow"])
        , Action = sb.optional_spec(sb.listof(sb.string_choice_spec(choices=["Update:Modify", "Update:Replace", "Update:Delete", "Update:*"])))
        , NotAction = sb.optional_spec(sb.listof(sb.string_choice_spec(choices=["Update:Modify", "Update:Replace", "Update:Delete", "Update:*"])))
        , Principal = sb.valid_string_spec(validators.regexed("^\*$"))
        , Resource = sb.optional_spec(sb.listof(sb.valid_string_spec(validators.regexed(r"^(LogicalResourceId/.*|\*)$"))))
        , NotResource = sb.optional_spec(sb.listof(sb.valid_string_spec(validators.regexed(r"^(LogicalResourceId/.*|\*)$"))))
        , Condition = sb.optional_spec(sb.dictionary_spec())
      ))
    )



"""
Responsible for finding tasks in the configuration and executing them
"""

from bespin.actions import available_actions, default_actions
from bespin.option_spec.task_objs import Task
from bespin.errors import BadTask

class TaskFinder(object):
    def __init__(self, collector):
        self.tasks = {}
        self.collector = collector

    def stack_finder(self, task):
        return getattr(self.tasks[task], "stack", self.collector.configuration['bespin'].chosen_stack)

    def task_runner(self, task, **kwargs):
        if task not in self.tasks:
            raise BadTask("Unknown task", task=task, available=self.tasks.keys())
        return self.tasks[task].run(self.collector, self.stack_finder(task), available_actions, self.tasks, **kwargs)

    def default_tasks(self):
        """Return default tasks"""
        return dict((name, Task(action=name, label="Bespin")) for name in default_actions)

    def find_tasks(self, overrides):
        """Find the custom tasks and record the associated stack with each task"""
        tasks = self.default_tasks()
        configuration = self.collector.configuration

        for stack in list(configuration["stacks"]):
            path = configuration.path(["stacks", stack, "tasks"], joined="stacks.{0}.tasks".format(stack))
            nxt = configuration.get(path, {})
            tasks.update(nxt)

        if overrides:
            tasks.update(overrides)

        self.tasks = tasks




from bespin.helpers import memoized_property

from boto.kms.layer1 import KMSConnection
from boto.regioninfo import get_regions

from input_algorithms.spec_base import NotSpecified
import logging
import base64
import json
import six

log = logging.getLogger("bespin.amazon.kms")

class KMSFixed(KMSConnection):
    def encrypt(self, key_id, plaintext, encryption_context=None, grant_tokens=None):
        """
        Borrowed from boto sourcecode, waiting for https://github.com/boto/boto/pull/3051
        """
        if not isinstance(plaintext, six.binary_type):
            raise TypeError(
                "Value of argument ``plaintext`` "
                "must be of type %s." % six.binary_type)
        plaintext = base64.b64encode(plaintext)
        params = {'KeyId': key_id, 'Plaintext': plaintext.decode('utf-8'), }
        if encryption_context is not None:
            params['EncryptionContext'] = encryption_context
        if grant_tokens is not None:
            params['GrantTokens'] = grant_tokens
        response = self.make_request(action='Encrypt',
                                        body=json.dumps(params))
        if response.get('CiphertextBlob') is not None:
            response['CiphertextBlob'] = base64.b64decode(
                response['CiphertextBlob'].encode('utf-8'))
        return response

    def decrypt(self, ciphertext_blob, encryption_context=None,
                grant_tokens=None):
        """
        Borrowed from boto sourcecode, waiting for https://github.com/boto/boto/pull/3051
        """
        if not isinstance(ciphertext_blob, six.binary_type):
            raise TypeError(
                "Value of argument ``ciphertext_blob`` "
                "must be of type %s." % six.binary_type)
        ciphertext_blob = base64.b64encode(ciphertext_blob)
        params = {'CiphertextBlob': ciphertext_blob.decode('utf-8'), }
        if encryption_context is not None:
            params['EncryptionContext'] = encryption_context
        if grant_tokens is not None:
            params['GrantTokens'] = grant_tokens
        response = self.make_request(action='Decrypt',
                                     body=json.dumps(params))
        if response.get('Plaintext') is not None:
            response['Plaintext'] = base64.b64decode(
                response['Plaintext'].encode('utf-8'))
        return response

class KMS(object):
    def __init__(self, region="ap-southeast-2"):
        self.region = region

    @memoized_property
    def conn(self):
        log.info("Using region [%s] for kms", self.region)
        for region in get_regions('kms', connection_cls=KMSFixed):
            if region.name == self.region:
                return region.connect()

    def decrypt(self, crypto_text, encryption_context=None, grant_tokens=None):
        if encryption_context is NotSpecified:
            encryption_context = None
        if grant_tokens is NotSpecified:
            grant_tokens = None
        return self.conn.decrypt(crypto_text, encryption_context, grant_tokens)

    def encrypt(self, key_id, plain_text, encryption_context=None, grant_tokens=None):
        if encryption_context is NotSpecified:
            encryption_context = None
        if grant_tokens is NotSpecified:
            grant_tokens = None
        return self.conn.encrypt(key_id, plain_text, encryption_context, grant_tokens)




from bespin.errors import StackDoesntExist, BadStack, Throttled
from bespin.amazon.mixin import AmazonMixin
from bespin import helpers as hp

import botocore
import boto3
import datetime
import logging
import pytz
import time
import six
import os

log = logging.getLogger("bespin.amazon.cloudformation")

class StatusMeta(object):
    def __new__(cls, name, bases, attrs):
        attrs["name"] = name
        attrs["failed"] = name.endswith("FAILED")
        attrs["complete"] = name.endswith("COMPLETE")
        attrs["in_progress"] = name.endswith("IN_PROGRESS")
        attrs["cleanup_in_progress"] = name.endswith("CLEANUP_IN_PROGRESS")

        attrs["is_create"] = name.startswith("CREATE")
        attrs["is_delete"] = name.startswith("DELETE")
        attrs["is_update"] = name.startswith("UPDATE") and not name.startswith("UPDATE_ROLLBACK")
        attrs["is_rollback"] = name.startswith("ROLLBACK") or name.startswith("UPDATE_ROLLBACK")
        return type(name, bases, attrs)

class Status(object):
    exists = True
    statuses = {}

    @classmethod
    def find(kls, name):
        if name in kls.statuses:
            return kls.statuses[name]
        return six.add_metaclass(StatusMeta)(type(name, (Status, ), {}))

class NONEXISTANT(Status):
    exists = False

class CREATE_IN_PROGRESS(Status): pass
class CREATE_FAILED(Status): pass
class CREATE_COMPLETE(Status): pass

class ROLLBACK_IN_PROGRESS(Status): pass
class ROLLBACK_FAILED(Status): pass
class ROLLBACK_COMPLETE(Status): pass

class DELETE_IN_PROGRESS(Status): pass
class DELETE_FAILED(Status): pass
class DELETE_COMPLETE(Status): pass

class UPDATE_IN_PROGRESS(Status): pass
class UPDATE_COMPLETE_CLEANUP_IN_PROGRESS(Status): pass
class UPDATE_COMPLETE(Status): pass
class UPDATE_ROLLBACK_IN_PROGRESS(Status): pass
class UPDATE_ROLLBACK_FAILED(Status): pass
class UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS(Status): pass
class UPDATE_ROLLBACK_COMPLETE(Status): pass

# REVIEW_IN_PROGRESS only valid for CreateChangeSet with ChangeSetType=CREATE
class REVIEW_IN_PROGRESS(Status): pass

for kls in [Status] + Status.__subclasses__():
    with_meta = six.add_metaclass(StatusMeta)(kls)
    locals()[kls.__name__] = with_meta
    Status.statuses[kls.__name__] = with_meta

##BOTO3 TODO: refactor to use boto3 resources
class Cloudformation(AmazonMixin):
    def __init__(self, stack_name, region="ap-southeast-2"):
        self.region = region
        self.stack_name = stack_name

    @hp.memoized_property
    def conn(self):
        log.info("Using region [%s] for cloudformation (%s)", self.region, self.stack_name)
        return self.session.client('cloudformation', region_name=self.region)

    @hp.memoized_property
    def session(self):
        return boto3.session.Session(region_name=self.region)

    def reset(self):
        self._description = None

    def description(self, force=False):
        """Get the descriptions for the stack"""
        if not getattr(self, "_description", None) or force:
            with self.catch_boto_400(StackDoesntExist, "Couldn't find stack"):
                while True:
                    try:
                        with self.ignore_throttling_error():
                            response = self.conn.describe_stacks(StackName=self.stack_name)
                            self._description = response['Stacks'][0]
                            break
                    except Throttled:
                        log.info("Was throttled, waiting a bit")
                        time.sleep(0.5)
        return self._description

    @property
    def outputs(self):
        self.wait()
        description = self.description()
        if 'Outputs' in description:
            return dict((out['OutputKey'], out['OutputValue']) for out in description['Outputs'])
        else:
            return {}

    @property
    def status(self):
        force = False
        last_status = getattr(self, "_last_status", None)
        if last_status is None:
            self._last_status = datetime.datetime.now()
            force = True
        else:
            if self._last_status + datetime.timedelta(seconds=3) < datetime.datetime.now():
                force = True
                self._last_status = None

        try:
            description = self.description(force=force)
            return Status.find(description['StackStatus'])
        except StackDoesntExist:
            return NONEXISTANT

    def map_logical_to_physical_resource_id(self, logical_id):
        response = self.conn.describe_stack_resource(StackName=self.stack_name, LogicalResourceId=logical_id)
        return response['StackResourceDetail']["PhysicalResourceId"]

    def _convert_tags(self, tags):
        """ helper to convert python dictionary into list of AWS Tag dicts """
        return [{'Key': k, 'Value': v} for k,v in tags.items()] if tags else None

    def create(self, stack, params, tags=None, policy=None, role_arn=None):
        log.info("Creating stack (%s)\ttags=%s", self.stack_name, tags)
        stack_tags = self._convert_tags(tags)
        stack_args = {
              'StackName': self.stack_name
            , 'TemplateBody': stack
            , 'Parameters': params
            , 'Capabilities': ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM']
            , 'DisableRollback': os.environ.get("DISABLE_ROLLBACK", 0) == "1"
        }
        if stack_tags: stack_args['Tags'] = stack_tags
        if policy: stack_args['StackPolicyBody'] = policy
        if role_arn: stack_args['RoleARN'] = role_arn
        self.conn.create_stack(**stack_args)
        return True

    def update(self, stack, params, tags=None, policy=None, role_arn=None):
        log.info("Updating stack (%s)\ttags=%s", self.stack_name, tags)
        stack_tags = self._convert_tags(tags)
        stack_args = {
              'StackName': self.stack_name
            , 'TemplateBody': stack
            , 'Parameters': params
            , 'Capabilities': ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM']
            # NOTE: DisableRollback is not supported by UpdateStack. It is a property of the stack that can only be set during stack creation
        }
        if stack_tags: stack_args['Tags'] = stack_tags
        if policy: stack_args['StackPolicyBody'] = policy
        if role_arn: stack_args['RoleARN'] = role_arn
        with self.catch_boto_400(BadStack, "Couldn't update the stack", stack_name=self.stack_name):
            try:
                self.conn.update_stack(**stack_args)
            except botocore.exceptions.ClientError as error:
                if error.response['Error']['Message'] == "No updates are to be performed.":
                    log.info("No updates were necessary!")
                    return False
                else:
                    raise
        return True

    def validate_template(self, filename):
        with self.catch_boto_400(BadStack, "Amazon says no", stack_name=self.stack_name, filename=filename):
            self.conn.validate_template(TemplateBody=open(filename).read())

    ##BOTO3 TODO: can this be refactored with client.get_waiter?
    ##BOTO3 TODO: also consider client.get_paginator('describe_stack_events')
    def wait(self, timeout=1200, rollback_is_failure=False, may_not_exist=True):
        status = self.status
        if not status.exists and may_not_exist:
            return status

        last = datetime.datetime.now(pytz.utc)
        if status.failed:
            raise BadStack("Stack is in a failed state, it must be deleted first", name=self.stack_name, status=status)

        for _ in hp.until(timeout, step=15):
            if status.exists and status.complete:
                break

            log.info("Waiting for %s - %s", self.stack_name, status.name)
            if status.exists and not status.complete:
                status = self.status
            else:
                break

            description = self.description()

            events = []
            while True:
                try:
                    with self.ignore_throttling_error():
                        response = self.conn.describe_stack_events(StackName=self.stack_name)
                        events = response['StackEvents']
                        break
                except Throttled:
                    log.info("Was throttled, waiting a bit")
                    time.sleep(1)

            next_last = events[0]['Timestamp']
            for event in events:
                if event['Timestamp'] > last:
                    reason = event.get('ResourceStatusReason', '')
                    log.info("%s - %s %s (%s) %s", self.stack_name, event['ResourceType'], event['LogicalResourceId'], event['ResourceStatus'], reason)
            last = next_last

        status = self.status
        if status.failed or (rollback_is_failure and status.is_rollback) or not status.complete:
            raise BadStack("Stack failed to complete", final_status=status)

        return status




from bespin.helpers import memoized_property, until
from bespin.errors import BadSQSMessage

import boto.sqs

from collections import namedtuple
import logging
import json

log = logging.getLogger("bespin.amazon.sqs")

class Message(namedtuple("Message", ["result", "instance_id", "output"])):
    @classmethod
    def decode(kls, message):
        """Takes a message that is : separated and maps it an instance of Message"""
        if message.count(':') < 2:
            raise BadSQSMessage("Less than two colons", msg=message)

        result, instance_id, output = message.split(':', 2)
        return kls(result=result, instance_id=instance_id, output=output)

class SQS(object):
    def __init__(self, region="ap-southeast-2"):
        self.region = region

    @memoized_property
    def conn(self):
        log.info("Using region [%s] for sqs", self.region)
        return boto.sqs.connect_to_region(self.region)

    def get_all_deployment_messages(self, sqs_url, timeout=60, sleep=2):
        """
        Get all the messages of the queue, dropping those that are not deployment messages

        We keep getting messages whilst the count is greater than 0 and we don't have messages yet.

        We will eventually timeout and return what we have if we keep getting invalid messages or keep getting no messages.
        """
        messages = []
        q = self.conn.get_queue(sqs_url)
        for _ in until(timeout, step=sleep):
            while q.count() > 0:
                raw_messages = self.conn.receive_message(q, number_messages=1)
                for raw_message in raw_messages:
                    encoded_message = json.loads(raw_message.get_body())['Message']

                    q.delete_message(raw_message)

                    try:
                        messages.append(Message.decode(encoded_message))
                    except BadSQSMessage as error:
                        log.error("Failed to parse a message: %s", error)

            if messages:
                break

        return messages




from setuptools import setup, find_packages
from bespin import VERSION

setup(
      name = "bespin"
    , version = VERSION
    , packages = ['bespin'] + ['bespin.%s' % pkg for pkg in find_packages('bespin')]
    , include_package_data = True

    , install_requires =
      [ "delfick_app==0.9.3"
      , "option_merge==1.5"
      , "input_algorithms==0.5.9"

      , "six"
      , "pytz"
      , "slacker"
      , "humanize"
      , "argparse"
      , "requests"
      , "paramiko==1.16.0"

      , "radssh==1.0.5"
      , "pyrelic==0.8.0"
      , "boto==2.40.0"
      , "boto3"
      , "pyYaml==3.10"
      , "ultra_rest_client==0.1.4"
      , "FileChunkIO==1.6"
      , "dnslib==0.9.4"
      ]

    , extras_require =
      { "tests":
        [ "noseOfYeti>=1.5.0"
        , "nose"
        , "mock"
        , "moto"

        # Need to ensure httpretty is not 0.8.7
        # To prevent an infinite loop in python3 tests
        , "httpretty==0.8.6"
        ]
      }

    , entry_points =
      { 'console_scripts' :
        [ 'bespin = bespin.executor:main'
        ]
      }

    # metadata for upload to PyPI
    , url = "https://github.com/delfick/bespin"
    , author = "Stephen Moore"
    , author_email = "delfick755@gmail.com"
    , description = "Opinionated wrapper around boto that reads yaml"
    , license = "MIT"
    , keywords = "cloudformation boto"
    )

